## two dynamic variables
handler.stack <- NULL  ## place holder for .First.lib
callcc <- function(fun) {
    env <- environment()
    k <- function(v, restarts.honored = FALSE) {
        if (restarts.honored)
            .Call("DoReturnOrRestart", v, env)
        else
            .Call("DoReturn", v, env)
    }
    fun(k)
}
add.to.handler.stack <- function(handler, class, exit, stack) {
    list(handler = handler, class = class, exit = exit,
         next.handler = stack)
}
add.default.handler <- function(handler, class)
    handler.stack(add.to.handler.stack(handler, class, NULL, handler.stack()))
handle.condition <- function(c) {
    h <- handler.stack()
    if (is.null(h))
         FailsafeErrorHandler(c)
    while (! is.null(h))
        if (inherits(c, h$class))
            break
        else h <- h$next.handler
    if (is.null(h)) {
        EnableErrorHooks()
        my.stop(no.condition.handler.exception(c)) #****
    }
    if (is.null(h$exit))
        dynamic.bind({
            EnableErrorHooks()
            h$handler(c)
        }, handler.stack = h$next.handler)
    else {
        restarts.honored <- inherits(c, "exception")
        result <- list(throw = TRUE, handler = h$handler, condition = c)
        EnableErrorHooks()
        h$exit(result, restarts.honored)
    }
}
FailsafeErrorHandler <- function(c) {
    errcat("Error: error in default exception handler\n")
    EnableErrorHooks()
    abort(TRUE)
}
signal.condition <- function(c) {
    if (! inherits(c, "condition"))
        c <- simple.condition(c)
    handle.condition(c)
}
my.stop <- function(e, call. = TRUE) {
    if (! is.condition(e))
        e <- simple.exception(e, if (call.) sys.call(1) else NULL)
    signal.condition(e)
    cat("aborting ...\n", file = stderr())
    abort(TRUE)
}
stop <- my.stop
my.warning <- function(w) {
    if (! inherits(w, "warning"))
       w <- simple.warning(w, sys.call(1))
    signal.condition(w)
}
warning <- my.warning
abort <- function(restarts.honored = FALSE)
    .Call("JumpToToplevel", restarts.honored)
ignore.errors <- function(expr)
    try.catch(expr, exception = function(e) e)
muffle.warnings <- function(expr)
    with.handlers(expr, warning = function(w) {})
with.handlers <- function(expr, ...) {
    stack <- handler.stack()
    handlers <- rev(list(...))
    classes <- names(handlers)
    for (i in seq(along = handlers))
        stack <- add.to.handler.stack(handlers[[i]], classes[i], NULL, stack)
    dynamic.bind(expr, handler.stack = stack)
}
try.catch <- function(expr, ..., finally = NULL) {
    on.exit(finally) #**** try(finally)???
    result <- callcc(function(k) {
        stack <- handler.stack()
        handlers <- rev(list(...))
        classes <- names(handlers)
        for (i in seq(along = handlers))
            stack <- add.to.handler.stack(handlers[[i]], classes[i], k, stack)
        dynamic.bind(list(throw = FALSE, value = expr), handler.stack = stack)
    })
    if (result$throw)
        result$handler(result$condition)
    else
        result$value
}
is.condition <- function(c) inherits(c, "condition")
print.condition <- function(c, ...) {
    msg <- condition.message(c)
    call <- condition.call(c)
    class <- class(c)[1]
    if (! is.null(call))
        cat("<", class, " in ", deparse(call), ": ", msg, ">\n", sep="")
    else
        cat("<", class, ": ", msg, ">\n", sep="")
}

condition.message <- function(c) UseMethod("condition.message", c)
condition.call <- function(c) UseMethod("condition.call", c)

condition.message.condition <- function(c) c$message
condition.call.condition <- function(c) c$call
simple.condition <- function(message, call = NULL) {
    class <- c("simple.condition", "condition")
    structure(list(message=as.character(message), call = call), class=class)
}

simple.exception <- function(message, call = NULL) {
    class <- c("simple.exception", "exception", "condition")
    structure(list(message=as.character(message), call = call), class=class)
}

simple.warning <- function(message, call = NULL) {
    class <- c("simple.warning", "warning", "condition")
    structure(list(message=as.character(message), call = call), class=class)
}
no.condition.handler.exception <- function(c)
    structure(list(message = paste("no condition handler for", class(c)[1]),
                   condition = c),
              class = c("no.condition.handler.exception",
                        "exception", "condition"))
default.handler <- function(e) {
    UseMethod("default.handler", e)
}
default.handler.exception <- function(e) {
    call <- condition.call(e)
    message <- condition.message(e)
    op <- getOption("add.error.underscore")
    if (is.null(op) || op)
        us <- "_"
    else
        us <- ""
    if (is.null(call))
        emsg <- paste(us, "Error: ", message, "\n", sep = "")
    else {
        dcall <- deparse(call)
        if (nchar(dcall) > 30)
            emsg <- paste(us, "Error in ", dcall, " :\n\t", message, "\n",
                          sep = "")
        else
            emsg <- paste(us, "Error in ", dcall, " : ", message, "\n",
                          sep = "")
    }

    seterrmessage(emsg)
    if (getOption("error.messages")) {
        errcat(emsg)
        PrintDeferredWarnings()
    }

    handler <- getOption("error")
    if (! is.null(handler))
        eval(handler, R_GlobalEnv)
    else if (! interactive()) {
        errcat("Execution halted\n")
        q("no", 1, FALSE)  # quit, no save, no .Last, status=1
    }
    #**** R_SinkReset(); ????

    tb <- getTraceback()
    tb <- trim.traceback(tb)
    assign(".Traceback", tb, env = .GlobalEnv)
    abort(TRUE)
}
trim.traceback <- function(t) {
    n <- length(t)
    pos <- NULL
    for (i in seq(along=t))
        if (pmatch("signal.condition(", t[[i]], 0)) {
            pos <- i
            break
        }
    if (is.null(pos))
        t
    else {
        if (pos < n - 1 &&
            pmatch("my.stop(", t[[pos + 1]], 0) &&
            pmatch("error.hook(", t[[pos + 2]], 0))
            pos <- pos + 2
        if (pos == n)
            NULL
        else
            t[(pos+1):n]
    }
}
PrintDeferredWarnings <- function() .Call("PrintDeferredWarnings")
getTraceback <- function(skip = 1)
    .Call("GetTraceback", as.integer(skip))
seterrmessage <- function(s)
    .C("SetErrmessage", as.character(s))
default.handler.warning <- function(w) {
    .Call("InternalWarningCall", condition.call(w), condition.message(w))
}
EnableErrorHooks <- function() {
    .Call("EnableExceptionHooks")
}
errcat<- function(s) cat(s, file=stderr())
error.hook <- function(call, msg)
     my.stop(simple.exception(msg, call))

warning.hook <- function(call, msg)
     my.warning(simple.warning(msg, call))
.First.lib <- function(lib, pkg) {
    library.dynam(pkg, pkg, lib)

    require(dynvars)
    handler.stack <<- dynamic.variable()
    add.default.handler(default.handler, "exception")
    add.default.handler(default.handler, "warning")
    EnableErrorHooks()    
}
